#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

//Changes:
//1. If not connected, is_thick=0 and we're done...
//2. It was segfaulting consistently.  Adding k=0 in the appropriate spot in Union seemed to fix it.

//Changes to comments:
//1) The "initial" elements of Thick are not K_{2,2}s.  [Indeed, what happens is not what I expected when writing that...]
//2) I temporarily removed this sentence as I don't understand it.  Thick unions do not quite do this, but they do
//account for all edges when computing the diameter of the intersection, so that the thick union is again a thick subgraph. [I mean: If P,Q are induced thick subgraphs, gluing them along their intersection does not yield an induced subgraph necessarily, but the subgraph induced by this union is thick.]
//3) If the first row of Thick consist of 1s only,
//4) Removed the following, as it might be the case that there's only one line: and Thick still has at least two nontrivial rows.
//5) regarding the question, the sizing of Cones[i] is correct, I think. For example, for i=1 you have to consider one vertex, namely the 0th one.
//6) slightly modified the comment to int s=Squares(n). I think we can explain what first generation means later.
//7) Replaced "there's no need to make individual elements of Thick larger than they need to be" with "experimentally this seems faster". It's more honest :P
//8) Thick now never has empty lines, I took your suggestion.
//9) Doesn't it take too long to manually enter a matrix? How about reading a list of edges? Then one could save a matrix in some file and feed it to the program with "cat file| erdos a n 0" [To check a specific matrix, I have to enter it somewhere, somehow.  A list of edges is of course faster than the entire matrix.  I just wanted to check a square, a pentagon, a small join, and a thick of order 1 graph.  If we want a serious Manual, it should either read the list of edges from a file or get the list of edges from stdin.]
//10) Speaking of which, in ErdosRenyi there's some commented code that I use for testing.
//11) I created a "verbose mode", that can be activated with -v. It's more convenient for debugging and figuring out which parts of the code are fast.
//12) The non-verbose mode outputs two 0-1 digits, one for connectedness and one for thickness. Feel free to deactivate the former if inconvenient!
//13) If convenient, we could remove the line breaks in non-verbose mode.
//14) Made sure that the number of lines is 666 :P

using std::vector;

//THE CODE WITH SOME EXPLANATION

//This algorithm operates on a matrix Adj, which is the adjacency matrix of a graph G.

vector<vector<char> > Adj;

//The goal is to find out whether or not G belongs to the smallest class T of graphs
//with the following properties:
//
//1. T contains K_{2,2}.
//2. T is closed under the operation of coning off non-complete subgraphs.
//3. T is closed under the operation of thick unions.

//The algorithm proceeds by first finding a collection of subgraphs of G that belong to T
//which are collected in the list Thick.  Thick is a list of n-vectors of 1s and 0s
//encoding which vertices belong to the (induced) subgraph.

vector<vector<char> > Thick;

//Thick is modified by replacing pairs
//of subgraphs whose intersection is nonempty and has diameter more than 1 by the single graph consisting
//of their union ("thick union") and by coning off non-complete subgraphs of elements of Thick.  The coning
//operation replaces induced subgraphs with induced subgraphs.
//  If the first row of Thick consist of 1s only, after performing
//the operations, then G is thick.  If this does not eventually happen, we reach a point where there
//are no further operations possible and we're relatively hyperbolic.

//The matrix Cones stores, for each non-edge of G, all vertices adjacent to both vertices of the non-edge.

vector<vector<vector<short> > > Cones;

//The Matrix Neigh stores, for each vertex i, the all vertices adjacent to i.

vector<vector<short> > Neigh;


//The functions used are as follows:

//1. The function

void ErdosRenyi(float p, int n);

//is used to produce the adjacency matrix Adj: beginning with a set of n vertices, ErdosRenyi
//decides independently with probability p whether each pair of vertices is adjacent.

//2. The function Squares is used to populate Thick initially.

int Squares(int n, int v);

//3. Union performs thick unions and conings of elements of Thick

int Union(int n, int s);

//4. CheckThick checks, after no more operations are possible, whether or not G is a thick graph
//by examining the first line of Thick.

bool CheckThick(int n);

//5. FindCones adds, for each pair j,k of nonadjacent vertices, each vertex i adjacent to both of j,k.

void FindCones(int n);

//6. Connected checks if the graph is connected

bool Connected(int n);

//7. Neighbors populates Neigh

void Neighbors(int n);


//*****************************************************************************************************
//*************************************   MAIN   ******************************************************
//*****************************************************************************************************

main(int argc, char* argv[])
{//The arguments come from the command line and are a and n (a is ignored in manual tests) and rand_man. Th option -v activates the verbose mode.

	int v=0;
	if (argv[1][0]=='-')
	{
		if (argv[1][1]=='v')
			v=1;
	}

        /*int rand_man=0;
        if (argv[2][0]=='-')
        {
                if (argv[2][1]=='m');
                        rand_man=1;
        }*/

//p is the density of edges, and p = a*log(n)/n, where n is the number of vertices:

	float p;

	float a=atof(argv[v+1]);

	int n=atoi(argv[v+2]);

	p=a*log(n)/n; //THIS IS THE DENSITY. 

	if (v==1)
		printf("p=%f\n", p);

//We now start actually doing stuff.  First, we make the nxn matrix Adj:

	Adj.resize(n);
	for (int i = 0; i < n; ++i)
		Adj[i].resize(n);

//and fill it with 0s and 1s randomly (maintaining symmetry!)

   ErdosRenyi(p,n);
      
	if (v==1)
		printf("Matrix generated\n");

//We now populate Neigh...

	Neighbors(n);

//...and check if the graph is connected.

        int is_thick=0;

	if (v==1)
	{
		if (Connected(n))
			printf("Connected\n");
		else
                    {
                        printf("Not connected\n");
                        printf("Not thick, but we'll find some thick subgraphs.\n");
                    }
	} else
		{
		if (Connected(n)==false)
			//printf("1");
                    {
                        //printf("0");
                        printf("%d\n", is_thick);
                        return 0;
                    }

		}
//We expect something like n^4p^4(1-p)^2 empty squares, but it's much less clear how these interact.  That's (more or less)
//an expected upper bound for the number of thick subgraphs.  However, we can just put 10 here, since Thick will be sized dynamically.

//We similarly size the matrix Cones:

	Cones.resize(n);
	for (int i = 0; i < n; ++i)
		Cones[i].resize(i);

//Cones is now effectively an (nxn) matrix with all entries above the main diagonal 0.  Each entry will be its own vector recording
//the vertices adjacent to both of the first two coordinates, if they are nonadjacent.

//Populate Cones:

	FindCones(n);

	if (v==1)
		printf("Cones found\n");


//The first _serious_ step is to compute the first generation thick subgraphs, s is their number:

	int s=Squares(n,v);

	if (v==1)
		printf("There are %d first generation thick subgraphs\n",s);

//Do the find thick/cone/find another/union procedure as long as it was just done successfully, and shrink Thick.  Stop when Union fails to do anything or when Thick only has one line.
//At that point, after shrinking, we're thick if and only if Thick has a single nonzer0 line (consisting entirely of 1s).

	bool u;

	int unioncount=0;

	do
	{
		u=false;
		if (Union(n,s)>0)
			u=true;
		s=Thick.size();
		unioncount++;
	} while (u);

	if (v==1)
		printf("There have been %d union/coning rounds\n", unioncount);


//Finally, determine if the final Thick is a single 1-vector.  If so, output 1; 0 otherwise.  The variable is_thick
//is printed to stdout when we're done; it's 1 or 0 according to
//whether or not we are thick.  The script that's calling main will be keeping track of how many
//tests resulted in thick (at given value of a, say) using this value.

	if (v==1)
	{
	if (CheckThick(n))
		printf("Thick\n");
	else
		printf("Relatively hyperbolic\n");
	}
	else
		{


		if (CheckThick(n))
			is_thick=1;
		else
			is_thick=0;

		printf("%d\n", is_thick);
		}
//return is_thick;
}

//***********************************************************************************************************************************
//********************************************   ErdosRenyi   ***********************************************************************
//***********************************************************************************************************************************

void ErdosRenyi(float p, int n)

{//This function sets Adj[i][j]=Adj[j][i] to 0 or 1 according to whether a random number r in [0,1] exceeds p.
	float r;
 	srand (time(NULL));

	for (int i=0;i<n;i++)
        {
		for (int j=0;j<i;j++)
		{
			r=(float)rand() / (float)RAND_MAX;
			if(r<p)
	        	{
				Adj[i][j]=1;
				Adj[j][i]=1;
			}
		}
	}


//1- Square
	/*Adj[0][1]=1;
	Adj[1][2]=1;
	Adj[2][3]=1;
	Adj[3][0]=1;
//2- Path of length 3 connecting opposite vertices of the square (comment out for the options below)
	/*Adj[0][4]=1;
	Adj[4][5]=1;
	Ajd[5][2]=1;
//3- K_{3,2} (keep 2 commented)

	Adj[0][4]=1;
	Adj[2][4]=1;

//4- cone on K_{3,2} (2 commented)

	Adj[1][5]=1;
	Adj[3][5]=1;


//5- thick union of two 4's (2 commented)

	Adj[6][7]=1;
	Adj[7][8]=1;
	Adj[8][9]=1;
	Adj[9][6]=1;
	Adj[6][4]=1;
	Adj[8][4]=1;
	Adj[7][5]=1;
	Adj[9][5]=1;

//6- vertex of valence 1

	Adj[10][0]=1;


//7- path of length 3
	Adj[10][11]=1;
	Adj[11][6]=1;


//Symmetryzer
	for (int i=0;i<n;i++)
	{
		for (int j=0;j<n;j++)
		{
		if (Adj[i][j]==1)
			Adj[j][i]=1;
		}
	}*/
}

//***********************************************************************************************************************************
//*********************************************   Neighbors   ***********************************************************************
//***********************************************************************************************************************************

void Neighbors(int n)
{//Neigh[i] contains the list of vertices adjacent to i
	Neigh.resize(n);

	for (int i=0; i<n; i++)
	{
		for (int j=0; j<n;j++)
		{
			if (Adj[i][j]==1)
			{
				Neigh[i].push_back(j);
			}
		}
	}
}

//***********************************************************************************************************************************
//********************************************   Squares   **************************************************************************
//***********************************************************************************************************************************

int Squares(int n, int v)

{//The input is the adjacency matrix Adj.  The output is a list Thick; each element of Thick is an n-element
//vector whose entries are 1,0 according to whether or not the corresponding vertex belongs to a certain induced subgraph.  In other words, Thick is a
//list of induced subgraphs.  Each element of Thick will be a thick subgraph. Thick does not store the list of all squares, this would be very memory-expensive. Instead, it groups them together performing thick unions.

//This function returns s, the number of lines in the list Thick.  s is at most the number of induced squares in the graph, but (when the density is
//sufficiently large) is considerably smaller.
	int s=0;

	int t;
	bool added;
	bool addedik;

	int connik;
	int connjl;
	int l;

	int coeff=1;


//i is the lowest index vertex of the square, we assume j<l to kill the symmetry
//the procedure is: Start from a vertex i, move to a neighbor j,
//move to a neighbor k, disregard k if it is adjacent to i, look for l completing the square

//Iterate over all pairs of adjacent vertices (i,j) (using Neigh)
	for (int i=0; i<n; i++)
	{
		if (s>1000*coeff)
		{
			if (v==1)
				printf("Overflow prevention, %d thick subgraphs\n",s);
			Union(n,s);
			s=Thick.size();
			coeff++;
			if (v==1)
				printf("Done, %d thick subgraphs\n",s);
		}

	for (int m=0; m<Neigh[i].size(); m++)
	{

		if (Neigh[i][m]>i)
		{
			int j=Neigh[i][m];
			for (int q=0; q<Neigh[j].size(); q++)
			{//Find a third vertex k adjacent to j but not i...
				if (Neigh[j][q]>i && Adj[i][Neigh[j][q]]==0)
				{//...and go through Cones[i][k] to find all vertices l, different from j,
                                //such that i,j,k,l is a square.  The variable connik is the total number of
                                //vertices we can add to make squares with the given diagonal, and each makes a new square
                                //in conjunction with j.  addedik is false as long as
					int k=Neigh[j][q];
					connik=Cones[k][i].size();
					addedik=false;

                                        //Go through all vertices adjacent to both i and k...
					for (int o=0; o<connik; o++)
					{
						l=Cones[k][i][o];

                                                //...if the vertex l is different from j, and not adjacent to j, then i,j,k,l is a square.
						if(j<l && Adj[j][l]==0)
						{
//adds i,j,k,l to the all subgraphs containing i,k or j,l, and if there are no such subgraphs, adds the square in a new line of Thick
							t=0;
							added=false;
							connjl=Cones[l][j].size();
							while (t<s)
							{//Move through the lines of Thick, and check if (i,k) belongs to the t^th (current) thick subgraph.
                                                        //If so, then add to this subgraph every vertex whose link contains (i,k).  The variable addedik ensures
                                                        //That this is only done once; experimentally this seems faster.
								if (!addedik && Thick[t][i]==1 && Thick[t][k]==1)
								{

									for (int q=0;q<connik;q++)
									{
										Thick[t][Cones[k][i][q]]=1;
									}
									addedik=true;
									added=true;

								}
								if (Thick[t][j]==1 && Thick[t][l]==1)
								{//Somewhat similarly, if the pair (j,l) -- the other diagonal of the square -- is contained in
                                                                //the t^th line of thick, then add each possible cone-point on (j,l) to the t^th line of Thick.
									for (int q=0;q<connjl;q++)
									{
										Thick[t][Cones[l][j][q]]=1;
									}
									added=true;
									connjl=0;
								}
								t++;
							}

//If nothing was added before, we still have a square with diagonals (i,k),(j,l), neither of which is contained in a preexisting thick subgraph.
//We therefore move through the vertices adjacent to both i,k and add them to the current line of Thick (this adds j and l, in particular) and then we
//move through the vertices adjacent to both j and l and add them to the line (this adds i,k).

							if (added==false)
							{
                                                        //we first add a line to Thick
								s++;
								Thick.resize(s);
								Thick[s-1].resize(n);
								for (int q=0;q<connik;q++)
								{
									Thick[s-1][Cones[k][i][q]]=1;
								}
								for (int q=0;q<connjl;q++)
								{
									Thick[s-1][Cones[l][j][q]]=1;
								}


							}
						}
					}
				}
			}
		}

	}
	}

	return s;
}

//***********************************************************************************************************************************
//********************************************   Union   ****************************************************************************
//***********************************************************************************************************************************

int Union(int n, int s)

{//The arguments are n (the length of a line in Thick, also the number of vertices) and s, the number of lines in Thick.  This function also knows about Adj and Thick, and will modify Thick as follows: for each non-edge (i,j), it finds a graph in Thick containing (i,j) and replaces it by the graph obtained by adding cone points.  It then looks for a second graph in Thick containing (i,j) and joins it to the (coned) first one.  If anything happens, we return a positive integer, which tells the loop in main to keep going...

	int first=-1;

//This function returns r, which is nonzero provided we were able to do at least one of the following: (1) find a Thick subgraph, and make it bigger by coning; (2) find a pair of Thick subgraphs and take their thick union.
	int r=0;

//The variable conn will hold, for a non-edge (i,j), the number of vertices k adjacent to i and j.
	int conn;

//iteration over all non-edges
	for (int i=0;i<n;i++)
	{
		int j=i+1;

		do
		{
			if (Adj[i][j]==0)
			{
//If (i,j) is not an edge, then we find a thick subgraph -- ie line in Thick -- that contains the non-edge (i,j).  The variable first is -1 as
//long as we have not found such a graph.  When we find such a graph, its position in Thick becomes first.  If we get to s, we're out of luck
//and move to next j.
				int k=-1;
				do
				{
					k++;
					if (Thick[k][i]==1 && Thick[k][j]==1)
						first=k;
				} while (first==-1 && k<s-1);

//Having found a line in Thick containing (i,j) (in fact, the first such line in Thick)...
				if (first !=-1)
				{//...make every possible cone on that line in Thick, i.e.
                                //add all vertices whose links in that subgraph have diameter at least 2.
                                //First take the size of the vector Cones[i][j] listing all vertices adjacent
                                //to both i and j...
					if (Cones[j][i].size()>0)
					{
						conn=Cones[j][i].size();
                                //For each vertex in Cones[i][j], add it to the relevant line in Thick.
						for (int l=0;l<conn;l++)
						{
							if (Thick[first][Cones[j][i][l]]==0)
							{
								Thick[first][Cones[j][i][l]]=1;
								r++;
							}
						}
//Uncomment the following line to make sure to cone over i,j just once
//						Cones[j][i].resize(0);
					}
                                //Note that if we were able to do any coning, the variable r became nonzero (and is never decreased).

					while (k<s-1)
					{
//Look for another subgraph in Thick containing (i,j) and, if one is found, make r nonzero, and combine this subgraph with Thick[first].
						k++;

						if (Thick[k][i]==1 && Thick[k][j]==1 && k<s)
						{
							r++;
							for (int l=0;l<n;l++)
							{

								if (Thick[first][l]==0 && Thick[k][l]==1)
									Thick[first][l]=1;

								Thick[k][l]=Thick[s-1][l];
							}
//not critical issue: k should be re-checked.
							s--;
						}
					}

					first=-1;
				}
			}
			j++;
		} while (j<n);
	}
	Thick.resize(s);
	return r;
}


//***********************************************************************************************************************************
//********************************************   CheckThick   ***********************************************************************
//***********************************************************************************************************************************

bool CheckThick(int n)

{//If the first row of Thick is all 1s, then we're thick, and this function returns true.  Otherwise, it returns false.

	int k=0;
	int j=0;

	int temp= Thick.size();
	//printf("%d\n", temp );
	if (Thick.size()>0)
	{
	do
	{
		if (Thick[0][j]==0)
			k=1;
			j++;
	} while (k==0 && j<n);
	}

	if (k==0)
		return true;
	else
		return false;

}

//***********************************************************************************************************************************
//********************************************   FindCones   ************************************************************************
//***********************************************************************************************************************************

void FindCones(int n)

{//If (j,k), for j<k, is a non-edge and i is adjacent to both j,k, then the vector Cones[k][j] has i added.

	for (int i=0; i<n; i++)
	{
		for (int m=0;m<Neigh[i].size();m++)
		{
			int j=Neigh[i][m];
			{
				for (int q=m+1;q<Neigh[i].size();q++)
				{
					int k=Neigh[i][q];
					if (Adj[j][k]==0)
					{
						Cones[k][j].push_back(i);
					}
				}
			}
		}
	}
}

//***********************************************************************************************************************************
//********************************************   Connected   ************************************************************************
//***********************************************************************************************************************************

bool Connected(int n)

{//Checks if the graph is connected

//a 1 in the vector comp corresponds to a vertex in the connected component of 0
	vector<int> comp;
	comp.resize(n);

//stores vertices in the connected component of 0 in the order they are found
	vector<int> compscramble;

	comp[0]=1;
	compscramble.push_back(0);

//the iteration adds to comp and compscramble all neighbors of the vertices in the connected component of 0 that have been already found
//goes on until no more vertices have been added

	int old;

	do
	{
		old=compscramble.size();
		for (int q=0; q<compscramble.size();q++)
		{
				int j=compscramble[q];
				for (int k=0;k<Neigh[j].size();k++)
				{
					if (comp[Neigh[j][k]]==0)
					{
						comp[Neigh[j][k]]=1;
						compscramble.push_back(Neigh[j][k]);
					}
				}
		}
	} while (old<compscramble.size());

//if the sixe of the connected component of 0 is n, then the graph is connected

	if (compscramble.size()==n)
		return true;
	else
		return false;
}

/*//***********************************************************************************************************************************
//********************************************   Manual   ***************************************************************************
//***********************************************************************************************************************************

void Manual(int n)

{//This just reads a symmetric matrix with 0s on the main diagonal from stdin.

    int entry = 0;

    for ( int i = 0; i<n; i++ )
    {
        for ( int j = 0; j<i; j++ )
        {
            printf ("Adj[%d][%d]=?\n", i, j);
            scanf ("%d", &entry);

            if ( entry==0 )
            {
                Adj[i][j]=0;
                Adj[j][i]=0;
            }
            else
            {
                Adj[i][j]=1;
                Adj[j][i]=1;
            }
        }
    }
}*/
